using CUDA
using .ExpressionProcessing
using .Interpreter

expressions = Vector{Expr}(undef, 2)
variables = Matrix{Float64}(undef, 2,2)
parameters = Vector{Vector{Float64}}(undef, 2)

# Resulting value should be 10 for the first expression
expressions[1] = :(x1 + 1 * x2 + p1)
expressions[2] = :(5 + x1 + 1 * x2 + p1 + p2)
variables[1,1] = 2.0
variables[2,1] = 3.0
variables[1,2] = 0.0
variables[2,2] = 5.0
parameters[1] = Vector{Float64}(undef, 1)
parameters[2] = Vector{Float64}(undef, 2)
parameters[1][1] = 5.0
parameters[2][1] = 5.0
parameters[2][2] = 0.0

function testHelper(expression::Expr, variables::Matrix{Float64}, parameters::Vector{Vector{Float64}}, expectedResult::Float64)
	postfix = Vector([expr_to_postfix(expression)])
	println(postfix)
	result = Interpreter.interpret(postfix, variables, parameters)
	@test isequal(result[1,1], expectedResult)
end

@testset "Test TMP interpretation" begin
	postfixExpr = expr_to_postfix(expressions[1])
	postfixExprs = Vector([postfixExpr])
	push!(postfixExprs, expr_to_postfix(expressions[2]))

	# CUDA.@sync interpret(postfixExprs, variables, parameters)
end

@testset "Test conversion to matrix" begin
	reference = Matrix{Float64}(undef, 2, 2)
	reference[1,1] = 5.0
	reference[2,1] = NaN64
	reference[1,2] = 5.0
	reference[2,2] = 0.0
	# reference = Matrix([5.0, NaN],
	# 				   [5.0, 0.0])
	CUDA.@sync result = Interpreter.convert_to_matrix(parameters, NaN64)

	@test isequal(result, reference)
end

@testset "Test commutative interpretation" begin
	var = Matrix{Float64}(undef, 2, 1)
	param = Vector{Vector{Float64}}(undef, 1)
	expectedResult = 8.0 # Not using eval because the variables are not stored in global scope
	
	var[1,1] = 3.0
	var[2,1] = 5.0
	param[1] = [3.0, 5.0]

	# test with fixed values
	expr = :(3.0 + 5.0)
	testHelper(expr, var, param, expectedResult)
	# test with variables
	expr = :(x1 + x2)
	testHelper(expr, var, param, expectedResult)
	# test with parameters
	expr = :(p1 + p2)
	testHelper(expr, var, param, expectedResult)
end

@testset "Test non commutative interpretation" begin
	var = Matrix{Float64}(undef, 2, 1)
	param = Vector{Vector{Float64}}(undef, 1)
	expectedResult = -2.0 # Not using eval because the variables are not stored in global scope
	
	var[1,1] = 3.0
	var[2,1] = 5.0
	param[1] = [3.0, 5.0]

	# test with fixed values
	expr = :(3.0 - 5.0)
	testHelper(expr, var, param, expectedResult)
	# One Test with fixed values but swapped
	expr = :(5.0 - 3.0)
	testHelper(expr, var, param, -expectedResult)
	
	# test with variables
	expr = :(x1 - x2)
	testHelper(expr, var, param, expectedResult)
	# test with variables but swapped
	expr = :(x2 - x1)
	testHelper(expr, var, param, -expectedResult)
	
	# test with parameters
	expr = :(p1 - p2)
	testHelper(expr, var, param, expectedResult)
	expr = :(p2 - p1)
	testHelper(expr, var, param, -expectedResult)
end

@testset "Test single value operator interpretation" begin
	var = Matrix{Float64}(undef, 1, 1)
	param = Vector{Vector{Float64}}(undef, 1)
	expectedResult = 3.0 # Not using eval because the variables are not stored in global scope
	
	var[1,1] = -3.0
	param[1] = [-3.0]

	println("SINGLE VALUE")
	# test with fixed value
	expr = :(abs(-3.0))
	testHelper(expr, var, param, expectedResult)
	# test with variable
	expr = :(abs(x1))
	testHelper(expr, var, param, expectedResult)
	# test with parameter
	expr = :(abs(p1))
	testHelper(expr, var, param, expectedResult)
end

# TODO: Add several tests fo the mathematical expressions
#       And some more complicated expressions, with some only having variables, some only having parameters and some having both